from __future__ import annotations

import re

from coauthorcheck.models import Trailer

TRAILER_TOKEN_PATTERN = re.compile(r"^[A-Za-z0-9-]+: .+")


def extract_trailer_block(message: str) -> list[tuple[int, str]]:
    lines = message.splitlines()
    end = len(lines) - 1

    while end >= 0 and not lines[end].strip():
        end -= 1

    if end < 0:
        return []

    block: list[tuple[int, str]] = []
    found_trailer = False

    index = end
    while index >= 0:
        line = lines[index]

        if not line.strip():
            break

        is_continuation = bool(block) and line[:1].isspace()
        is_trailer = bool(TRAILER_TOKEN_PATTERN.match(line))

        if not found_trailer and not is_trailer:
            return []

        if is_trailer:
            found_trailer = True
            block.append((index + 1, line))
            index -= 1
            continue

        if is_continuation:
            block.append((index + 1, line))
            index -= 1
            continue

        break

    block.reverse()
    return block


def extract_trailers(message: str) -> list[Trailer]:
    trailers: list[Trailer] = []
    for line_number, line in extract_trailer_block(message):
        token, value = line.split(":", 1)
        trailers.append(
            Trailer(
                token=token.strip(),
                value=value.strip(),
                raw=line,
                line_number=line_number,
            )
        )
    return trailers


def extract_coauthor_trailers(message: str) -> list[Trailer]:
    return [
        trailer
        for trailer in extract_trailers(message)
        if trailer.token.lower() == "co-authored-by"
    ]
    


from __future__ import annotations

from dataclasses import dataclass, fields
from enum import StrEnum
from pathlib import Path
import tomllib


class ConfigError(ValueError):
    """Raised when configuration cannot be parsed."""


class Severity(StrEnum):
    ERROR = "error"
    WARNING = "warning"


def _normalize_rule_value(value: object) -> Severity | None:
    if value is None:
        return None
    if value is True:
        return Severity.ERROR
    if value is False:
        return None
    if isinstance(value, Severity):
        return value
    if isinstance(value, str):
        normalized = value.lower()
        if normalized == Severity.ERROR:
            return Severity.ERROR
        if normalized == Severity.WARNING:
            return Severity.WARNING
    raise ValueError(f"Invalid rule value: {value!r}")


@dataclass(slots=True)
class RuleConfig:
    github_handle: Severity | bool | str | None = Severity.ERROR
    incorrect_casing: Severity | bool | str | None = Severity.ERROR
    invalid_format: Severity | bool | str | None = Severity.ERROR
    malformed_email: Severity | bool | str | None = Severity.ERROR
    missing_email: Severity | bool | str | None = Severity.ERROR
    missing_name: Severity | bool | str | None = Severity.ERROR
    name_parts: Severity | bool | str | None = Severity.ERROR
    email_domain: Severity | bool | str | None = None

    def __post_init__(self) -> None:
        for field in fields(type(self)):
            normalized = _normalize_rule_value(getattr(self, field.name))
            setattr(self, field.name, normalized)

    def severity_for(self, rule_name: str) -> Severity | None:
        return getattr(self, rule_name)


@dataclass(slots=True)
class Config:
    rules: RuleConfig
    allowed_email_domains: tuple[str, ...] = ()
    blocked_email_domains: tuple[str, ...] = ()
    allow_github_noreply: bool | None = None
    ignore_bots: bool = False
    minimum_name_parts: int = 2


DEFAULT_CONFIG = Config(rules=RuleConfig())


def load_config(config_path: Path | None = None, start_dir: Path | None = None) -> Config:
    if config_path is not None:
        return _load_config_file(config_path)

    for directory in _iter_search_dirs(start_dir or Path.cwd()):
        dedicated_path = directory / ".coauthorcheck.toml"
        if dedicated_path.is_file():
            return _load_config_file(dedicated_path)

        pyproject_path = directory / "pyproject.toml"
        if pyproject_path.is_file():
            return _load_config_file(pyproject_path)

    return Config(rules=RuleConfig())


def _iter_search_dirs(start_dir: Path) -> list[Path]:
    resolved = start_dir.resolve()
    return [resolved, *resolved.parents]


def _load_config_file(path: Path) -> Config:
    resolved = path.resolve()
    data = tomllib.loads(resolved.read_text(encoding="utf-8"))

    if resolved.name == "pyproject.toml":
        config_data = data.get("tool", {}).get("coauthorcheck", {})
    else:
        config_data = data

    return _parse_config(config_data, resolved)


def _parse_config(data: dict, path: Path) -> Config:
    if not isinstance(data, dict):
        raise ConfigError(f"Invalid configuration in {path}.")

    rules_data = data.get("rules", {})
    if not isinstance(rules_data, dict):
        raise ConfigError(f"'rules' must be a table in {path}.")

    policy_data = data.get("policy", {})
    if not isinstance(policy_data, dict):
        raise ConfigError(f"'policy' must be a table in {path}.")

    allowed_rule_names = {field.name for field in fields(RuleConfig)}
    unknown_rule_names = sorted(set(rules_data) - allowed_rule_names)
    if unknown_rule_names:
        names = ", ".join(unknown_rule_names)
        raise ConfigError(f"Unknown rule setting(s) in {path}: {names}.")

    rule_values: dict[str, bool | str] = {}
    for name, value in rules_data.items():
        if not isinstance(value, (bool, str)):
            raise ConfigError(
                f"Rule setting '{name}' in {path} must be a boolean or one of: 'error', 'warning'."
            )
        if isinstance(value, str) and value.lower() not in {Severity.ERROR, Severity.WARNING}:
            raise ConfigError(
                f"Rule setting '{name}' in {path} must be a boolean or one of: 'error', 'warning'."
            )
        rule_values[name] = value

    try:
        allowed_email_domains = policy_data.get("allowed_email_domains", [])
        if not isinstance(allowed_email_domains, list):
            raise ConfigError(f"'policy.allowed_email_domains' in {path} must be an array of strings.")
        blocked_email_domains = policy_data.get("blocked_email_domains", [])
        if not isinstance(blocked_email_domains, list):
            raise ConfigError(f"'policy.blocked_email_domains' in {path} must be an array of strings.")
        allow_github_noreply = policy_data.get("allow_github_noreply")
        if allow_github_noreply is not None and not isinstance(allow_github_noreply, bool):
            raise ConfigError(f"'policy.allow_github_noreply' in {path} must be a boolean.")
        ignore_bots = policy_data.get("ignore_bots", False)
        if not isinstance(ignore_bots, bool):
            raise ConfigError(f"'policy.ignore_bots' in {path} must be a boolean.")

        normalized_domains: list[str] = []
        for domain in allowed_email_domains:
            if not isinstance(domain, str) or not domain.strip():
                raise ConfigError(f"'policy.allowed_email_domains' in {path} must contain non-empty strings.")
            normalized_domains.append(domain.strip().lower())

        normalized_blocked_domains: list[str] = []
        for domain in blocked_email_domains:
            if not isinstance(domain, str) or not domain.strip():
                raise ConfigError(f"'policy.blocked_email_domains' in {path} must contain non-empty strings.")
            normalized_blocked_domains.append(domain.strip().lower())

        minimum_name_parts = policy_data.get("minimum_name_parts", 2)
        if not isinstance(minimum_name_parts, int) or isinstance(minimum_name_parts, bool):
            raise ConfigError(f"'policy.minimum_name_parts' in {path} must be an integer.")
        if minimum_name_parts < 1:
            raise ConfigError(f"'policy.minimum_name_parts' in {path} must be greater than or equal to 1.")

        if (
            normalized_domains
            or normalized_blocked_domains
            or allow_github_noreply is not None
        ) and "email_domain" not in rule_values:
            rule_values["email_domain"] = Severity.ERROR

        email_domain_severity = _normalize_rule_value(rule_values.get("email_domain"))
        if (
            email_domain_severity is not None
            and not normalized_domains
            and not normalized_blocked_domains
            and allow_github_noreply is None
        ):
            raise ConfigError(
                f"'rules.email_domain' in {path} requires 'policy.allowed_email_domains' or "
                f"'policy.blocked_email_domains' or 'policy.allow_github_noreply' to be set."
            )

        return Config(
            rules=RuleConfig(**rule_values),
            allowed_email_domains=tuple(normalized_domains),
            blocked_email_domains=tuple(normalized_blocked_domains),
            allow_github_noreply=allow_github_noreply,
            ignore_bots=ignore_bots,
            minimum_name_parts=minimum_name_parts,
        )
    except ValueError as error:
        raise ConfigError(f"Invalid configuration in {path}: {error}") from error



from __future__ import annotations

import re

from coauthorcheck.config import Config, DEFAULT_CONFIG, Severity
from coauthorcheck.models import Trailer, ValidationIssue, ValidationResult
from coauthorcheck.parser import extract_coauthor_trailers

COAUTHOR_TOKEN = "Co-authored-by"
NAME_EMAIL_PATTERN = re.compile(r"^(?P<name>.*?)\s*<(?P<email>[^<>]+)>$")
EMAIL_PATTERN = re.compile(r"^[^@\s<>]+@[^@\s<>]+\.[^@\s<>]+$")
GITHUB_HANDLE_PATTERN = re.compile(r"^@[A-Za-z0-9-]+$")
EMBEDDED_EMAIL_PATTERN = re.compile(r"(?P<email>[^@\s<>]+@[^@\s<>]+\.[^@\s<>]+)$")
GITHUB_NOREPLY_DOMAINS = {"users.noreply.github.com", "noreply.github.com"}
BOT_NAME_PATTERN = re.compile(r"(?i)(?:^|[\s_-])[a-z0-9_.-]*bot(?:$|[\s_-])")
GITHUB_BOT_SUFFIX_PATTERN = re.compile(r"(?i)\[bot\]$")


def _build_issue(
    code: str,
    message: str,
    line_number: int,
    severity: Severity,
    suggestion: str | None = None,
) -> ValidationIssue:
    return ValidationIssue(
        code=code,
        message=message,
        line_number=line_number,
        severity=severity,
        suggestion=suggestion,
    )


def _suggested_trailer(name: str | None = None, email: str | None = None) -> str:
    suggested_name = name.strip() if name and name.strip() else "Full Name"
    suggested_email = email.strip() if email and email.strip() else "email@example.com"
    return f"{COAUTHOR_TOKEN}: {suggested_name} <{suggested_email}>"


def _extract_trailer_parts(trailer: Trailer) -> tuple[str | None, str | None]:
    match = NAME_EMAIL_PATTERN.match(trailer.value)
    if match is not None:
        name = match.group("name").strip() or None
        email = match.group("email").strip() or None
        return name, email

    raw_value = trailer.value.strip()
    name_part, separator, email_part = raw_value.partition("<")
    if separator:
        name = name_part.strip() or None
        email = email_part.replace(">", "").strip() or None
        return name, email

    embedded_email = EMBEDDED_EMAIL_PATTERN.search(raw_value)
    if embedded_email is not None:
        email = embedded_email.group("email").strip()
        name = raw_value[: embedded_email.start()].strip() or None
        return name, email

    return raw_value or None, None


def _merged_suggestion(trailer: Trailer, issue_codes: set[str], config: Config) -> str:
    name, email = _extract_trailer_parts(trailer)

    if "missing-name" in issue_codes or not name:
        name = "Full Name"
    elif "github-handle" in issue_codes:
        name = "Full Name"
    elif "name-parts" in issue_codes:
        current_parts = name.split() if name else []
        while len(current_parts) < config.minimum_name_parts:
            current_parts.append("Surname")
        name = " ".join(current_parts)

    if ("missing-email" in issue_codes and not email) or not email:
        email = "email@example.com"
    elif "malformed-email" in issue_codes:
        email = "email@example.com"
    elif "email-domain" in issue_codes:
        suggested_email = _suggest_email_for_allowed_domain(email, config.allowed_email_domains)
        if suggested_email is not None:
            email = suggested_email

    return _suggested_trailer(name=name, email=email)


def _suggest_email_for_allowed_domain(
    email: str | None,
    allowed_domains: tuple[str, ...],
) -> str | None:
    if not allowed_domains:
        return None

    local_part = "user"
    if email and "@" in email:
        candidate = email.split("@", 1)[0].strip()
        if candidate:
            local_part = candidate

    if len(allowed_domains) == 1:
        return f"{local_part}@{allowed_domains[0]}"

    return None


def _email_domain_issue(name: str, email: str, line_number: int, config: Config) -> ValidationIssue:
    suggestion = None
    suggested_email = _suggest_email_for_allowed_domain(email, config.allowed_email_domains)
    if suggested_email is not None:
        suggestion = _suggested_trailer(name=name, email=suggested_email)

    return _build_issue(
        "email-domain",
        "Use an email address from an allowed, non-blocked domain.",
        line_number,
        severity=config.rules.severity_for("email_domain"),
        suggestion=suggestion,
    )


def _is_github_noreply_domain(domain: str) -> bool:
    return domain in GITHUB_NOREPLY_DOMAINS


def _apply_suggestion(issues: list[ValidationIssue], trailer: Trailer, config: Config) -> list[ValidationIssue]:
    if not issues:
        return issues

    if not any(issue.suggestion for issue in issues):
        issue_codes = {issue.code for issue in issues}
        if issue_codes == {"email-domain"} and _suggest_email_for_allowed_domain(
            _extract_trailer_parts(trailer)[1],
            config.allowed_email_domains,
        ) is None:
            return issues

    suggestion = _merged_suggestion(trailer, {issue.code for issue in issues}, config)
    for issue in issues:
        issue.suggestion = suggestion
    return issues


def _name_part_issue(config: Config, line_number: int) -> ValidationIssue:
    minimum_name_parts = config.minimum_name_parts
    code = "name-parts"
    if minimum_name_parts == 2:
        message = "Use at least a first and last name in the trailer."
    else:
        message = f"Use at least {minimum_name_parts} name parts in the trailer."
    return _build_issue(
        code,
        message,
        line_number,
        severity=config.rules.severity_for("name_parts"),
    )


def _looks_like_bot_identity(name: str | None, email: str | None = None) -> bool:
    normalized_name = (name or "").strip()
    normalized_email = (email or "").strip().lower()

    if normalized_name:
        if GITHUB_BOT_SUFFIX_PATTERN.search(normalized_name):
            return True
        if BOT_NAME_PATTERN.search(normalized_name):
            return True

    if normalized_email:
        local_part, _, domain = normalized_email.partition("@")
        if GITHUB_BOT_SUFFIX_PATTERN.search(local_part):
            return True
        if BOT_NAME_PATTERN.search(local_part):
            return True
        if domain in GITHUB_NOREPLY_DOMAINS and local_part.endswith("[bot]"):
            return True

    return False


def validate_trailer(trailer: Trailer, config: Config = DEFAULT_CONFIG) -> list[ValidationIssue]:
    issues: list[ValidationIssue] = []
    match = NAME_EMAIL_PATTERN.match(trailer.value)

    trailer_name, trailer_email = _extract_trailer_parts(trailer)
    if config.ignore_bots and _looks_like_bot_identity(trailer_name, trailer_email):
        return []

    severity = config.rules.severity_for("incorrect_casing")
    if severity and trailer.token != COAUTHOR_TOKEN:
        issues.append(
            _build_issue(
                "incorrect-casing",
                f"Use the exact trailer token '{COAUTHOR_TOKEN}'.",
                trailer.line_number,
                severity=severity,
            )
        )

    if match is None:
        severity = config.rules.severity_for("invalid_format")
        if severity:
            issues.append(
                _build_issue(
                    "invalid-format",
                    "Trailer must use the format 'Co-authored-by: Full Name <email@example.com>'.",
                    trailer.line_number,
                    severity=severity,
                )
            )

        severity = config.rules.severity_for("missing_email")
        if severity and ("<" not in trailer.value or ">" not in trailer.value):
            issues.append(
                _build_issue(
                    "missing-email",
                    "Add an email address in angle brackets, for example <email@example.com>.",
                    trailer.line_number,
                    severity=severity,
                )
            )

        severity = config.rules.severity_for("missing_name")
        if severity and (
            trailer.value.startswith("<") or not trailer.value.split("<", 1)[0].strip()
        ):
            issues.append(
                _build_issue(
                    "missing-name",
                    "Add the author name before the email address.",
                    trailer.line_number,
                    severity=severity,
                )
            )

        return _apply_suggestion(_dedupe_issues(issues), trailer, config)

    name = match.group("name").strip()
    email = match.group("email").strip()

    if not name:
        severity = config.rules.severity_for("missing_name")
        if severity:
            issues.append(
                _build_issue(
                    "missing-name",
                    "Add the author name before the email address.",
                    trailer.line_number,
                    severity=severity,
                )
            )
        severity = config.rules.severity_for("invalid_format")
        if severity:
            issues.append(
                _build_issue(
                    "invalid-format",
                    "Trailer must use the format 'Co-authored-by: Full Name <email@example.com>'.",
                    trailer.line_number,
                    severity=severity,
                )
            )
    else:
        severity = config.rules.severity_for("github_handle")
        if severity and GITHUB_HANDLE_PATTERN.match(name):
            issues.append(
                _build_issue(
                    "github-handle",
                    "Use a full name instead of a GitHub handle in the trailer.",
                    trailer.line_number,
                    severity=severity,
                )
            )
        else:
            severity = config.rules.severity_for("name_parts")
            if severity and len(name.split()) < config.minimum_name_parts:
                issues.append(_name_part_issue(config, trailer.line_number))

    severity = config.rules.severity_for("missing_email")
    if severity and not email:
        issues.append(
            _build_issue(
                "missing-email",
                "Add an email address in angle brackets, for example <email@example.com>.",
                trailer.line_number,
                severity=severity,
            )
        )
    else:
        severity = config.rules.severity_for("malformed_email")
        if severity and not EMAIL_PATTERN.match(email):
            issues.append(
                _build_issue(
                    "malformed-email",
                    "Use a valid email address in angle brackets.",
                    trailer.line_number,
                    severity=severity,
                )
            )
        else:
            severity = config.rules.severity_for("email_domain")
            if severity and (
                config.allowed_email_domains
                or config.blocked_email_domains
                or config.allow_github_noreply is not None
            ):
                domain = email.rsplit("@", 1)[1].lower()
                is_github_noreply = _is_github_noreply_domain(domain)
                if is_github_noreply and config.allow_github_noreply is True:
                    violates_allowed_domains = False
                    violates_blocked_domains = False
                    violates_github_noreply = False
                else:
                    violates_allowed_domains = bool(
                        config.allowed_email_domains and domain not in config.allowed_email_domains
                    )
                    violates_blocked_domains = domain in config.blocked_email_domains
                    violates_github_noreply = is_github_noreply and config.allow_github_noreply is False
                if violates_allowed_domains or violates_blocked_domains or violates_github_noreply:
                    issues.append(_email_domain_issue(name, email, trailer.line_number, config))

    return _apply_suggestion(_dedupe_issues(issues), trailer, config)


def validate_message(
    source: str,
    message: str,
    config: Config = DEFAULT_CONFIG,
    author_name: str | None = None,
    author_email: str | None = None,
) -> ValidationResult:
    issues: list[ValidationIssue] = []

    if config.ignore_bots and _looks_like_bot_identity(author_name, author_email):
        return ValidationResult(source=source, issues=[])

    for trailer in extract_coauthor_trailers(message):
        issues.extend(validate_trailer(trailer, config=config))

    return ValidationResult(source=source, issues=_dedupe_issues(issues))


def _dedupe_issues(issues: list[ValidationIssue]) -> list[ValidationIssue]:
    seen: set[tuple[str, int]] = set()
    deduped: list[ValidationIssue] = []

    for issue in issues:
        key = (issue.code, issue.line_number)
        if key in seen:
            continue
        seen.add(key)
        deduped.append(issue)

    return deduped



from __future__ import annotations

from dataclasses import dataclass, field

from coauthorcheck.config import Severity


@dataclass(slots=True)
class CommitMessage:
    source: str
    message: str
    author_name: str | None = None
    author_email: str | None = None


@dataclass(slots=True)
class Trailer:
    token: str
    value: str
    raw: str
    line_number: int


@dataclass(slots=True)
class ValidationIssue:
    code: str
    message: str
    line_number: int
    severity: Severity = Severity.ERROR
    suggestion: str | None = None


@dataclass(slots=True)
class ValidationResult:
    source: str
    issues: list[ValidationIssue] = field(default_factory=list)

    @property
    def is_valid(self) -> bool:
        return self.error_count == 0

    @property
    def error_count(self) -> int:
        return sum(issue.severity == Severity.ERROR for issue in self.issues)

    @property
    def warning_count(self) -> int:
        return sum(issue.severity == Severity.WARNING for issue in self.issues)



from __future__ import annotations

import subprocess
from dataclasses import dataclass

from coauthorcheck.models import CommitMessage


@dataclass(slots=True)
class GitError(RuntimeError):
    """Raised when git cannot provide commit messages."""

    message: str
    hint: str | None = None
    raw_error: str | None = None

    def __str__(self) -> str:
        return self.message


def read_commit_message(commit: str) -> CommitMessage:
    return CommitMessage(
        source=commit,
        message=_run_git(["show", "-s", "--format=%B", commit]),
        author_name=_run_git(["show", "-s", "--format=%an", commit]).strip() or None,
        author_email=_run_git(["show", "-s", "--format=%ae", commit]).strip() or None,
    )


def read_commit_range(commit_range: str) -> list[CommitMessage]:
    output = _run_git(["rev-list", "--reverse", commit_range])
    commits = [line.strip() for line in output.splitlines() if line.strip()]
    return [read_commit_message(commit) for commit in commits]


def clean_commit_message_text(message: str, comment_char: str = "#") -> str:
    cleaned_lines: list[str] = []
    scissors_marker = f"{comment_char} ------------------------ >8 ------------------------"

    for line in message.splitlines():
        if line == scissors_marker:
            break
        if line.startswith(comment_char):
            continue
        cleaned_lines.append(line)

    cleaned_message = "\n".join(cleaned_lines).rstrip()
    if message.endswith("\n"):
        return f"{cleaned_message}\n" if cleaned_message else ""
    return cleaned_message


def _run_git(args: list[str]) -> str:
    completed = subprocess.run(
        ["git", *args],
        check=False,
        capture_output=True,
        text=True,
    )
    if completed.returncode != 0:
        stderr = completed.stderr.strip() or "git command failed"
        raise interpret_git_error(stderr, args)
    return completed.stdout


def interpret_git_error(stderr: str, args: list[str]) -> GitError:
    lowered = stderr.lower()
    revision = args[-1] if args else None

    if "not a git repository" in lowered:
        return GitError(
            message="current directory is not a git repository.",
            hint="Run this command inside a Git repository, or pass a commit message file path instead.",
            raw_error=stderr,
        )

    if "detected dubious ownership" in lowered or "safe.directory" in lowered:
        return GitError(
            message="git refused to read this repository because it is not marked as safe.",
            hint="Run 'git config --global --add safe.directory <repo-path>' for this repository.",
            raw_error=stderr,
        )

    if any(text in lowered for text in ("unknown revision", "bad revision", "ambiguous argument")):
        target = f" '{revision}'" if revision else ""
        if revision and ".." in revision:
            return GitError(
                message=f"unable to resolve revision range{target}.",
                hint=(
                    "Try a smaller range like 'HEAD~1..HEAD', or validate only commits introduced by "
                    "your branch with 'main..HEAD' or 'origin/main..HEAD'."
                ),
                raw_error=stderr,
            )

        return GitError(
            message=f"unable to resolve revision{target}.",
            hint="Check that the commit, branch, or tag exists in the current repository.",
            raw_error=stderr,
        )

    return GitError(
        message="git failed to read commit messages.",
        hint=stderr,
        raw_error=stderr,
    )



from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version as package_version
from enum import StrEnum
from pathlib import Path
from typing import Annotated, Sequence

import click
import typer
from rich.console import Console
from rich.table import Table

from coauthorcheck.config import ConfigError, Severity, load_config
from coauthorcheck.git_utils import GitError, clean_commit_message_text, read_commit_message, read_commit_range
from coauthorcheck.models import CommitMessage, ValidationResult
from coauthorcheck.validation import validate_message

app = typer.Typer(
    add_completion=False,
    help="Validate Co-authored-by trailers in commit messages.",
    no_args_is_help=False,
    pretty_exceptions_enable=False,
)
console = Console()
error_console = Console(stderr=True)


class OutputFormat(StrEnum):
    TEXT = "text"
    JSON = "json"


JSON_SCHEMA_VERSION = 1


def _tool_version() -> str:
    try:
        return package_version("coauthorcheck")
    except PackageNotFoundError:
        return "unknown"


def load_messages(
    input_value: str | None,
    file_path: Path | None,
    commit: str | None,
    commit_range: str | None,
) -> list[CommitMessage]:
    explicit_values = [value for value in (file_path, commit, commit_range) if value is not None]
    if input_value and explicit_values:
        raise ValueError("Provide either a positional input or an explicit flag, not both.")

    if file_path is not None:
        return [
            CommitMessage(
                source=str(file_path),
                message=clean_commit_message_text(file_path.read_text(encoding="utf-8")),
            )
        ]

    if commit is not None:
        return [read_commit_message(commit)]

    if commit_range is not None:
        return read_commit_range(commit_range)

    if input_value is None:
        raise ValueError("One input source is required.")

    detected_kind = detect_input_kind(input_value)
    if detected_kind == "file":
        path = Path(input_value)
        return [
            CommitMessage(
                source=str(path),
                message=clean_commit_message_text(path.read_text(encoding="utf-8")),
            )
        ]
    if detected_kind == "range":
        return read_commit_range(input_value)
    return [read_commit_message(input_value)]


def detect_input_kind(value: str) -> str:
    if Path(value).exists():
        return "file"
    if ".." in value:
        return "range"
    return "commit"


def render_result(result: ValidationResult) -> None:
    if not result.issues:
        console.print(f"[green]OK[/green] {result.source}")
        return

    label = "[red]Issues[/red]"
    if result.is_valid and result.warning_count > 0:
        label = "[yellow]Warnings[/yellow]"
    console.print(f"{label} {result.source} ({len(result.issues)})")
    has_suggestions = any(issue.suggestion for issue in result.issues)
    table = Table(show_header=True, header_style="bold")
    table.add_column("Line", style="cyan", justify="right", no_wrap=True)
    table.add_column("Severity", no_wrap=True)
    table.add_column("Code", style="magenta", no_wrap=True)
    table.add_column("Message", style="white")
    if has_suggestions:
        table.add_column("Suggestion", style="green")

    shown_suggestion_lines: set[int] = set()
    for issue in result.issues:
        severity = "error" if issue.severity == Severity.ERROR else "warning"
        row = [str(issue.line_number), severity, issue.code, issue.message]
        if has_suggestions:
            suggestion = ""
            if issue.suggestion and issue.line_number not in shown_suggestion_lines:
                suggestion = issue.suggestion
                shown_suggestion_lines.add(issue.line_number)
            row.append(suggestion)
        table.add_row(*row)

    console.print(table)


def render_summary(results: list[ValidationResult]) -> None:
    commit_count = len(results)
    invalid_count = sum(not result.is_valid for result in results)
    warning_commit_count = sum(result.warning_count > 0 and result.is_valid for result in results)
    issue_count = sum(len(result.issues) for result in results)
    error_count = sum(result.error_count for result in results)
    warning_count = sum(result.warning_count for result in results)
    status = "[green]PASS[/green]" if invalid_count == 0 else "[red]FAIL[/red]"
    console.print(
        f"Summary: {status} - checked {commit_count} commit(s), "
        f"{invalid_count} commit(s) with errors, "
        f"{warning_commit_count} commit(s) with warnings, "
        f"{issue_count} total issue(s) ({error_count} error(s), {warning_count} warning(s))"
    )


def _serialize_results(results: list[ValidationResult]) -> dict:
    commit_count = len(results)
    invalid_count = sum(not result.is_valid for result in results)
    issue_count = sum(len(result.issues) for result in results)
    error_count = sum(result.error_count for result in results)
    warning_count = sum(result.warning_count for result in results)
    warning_commit_count = sum(result.warning_count > 0 and result.is_valid for result in results)
    status = "pass" if invalid_count == 0 else "fail"
    return {
        "schema_version": JSON_SCHEMA_VERSION,
        "tool_version": _tool_version(),
        "status": status,
        "summary": {
            "commit_count": commit_count,
            "invalid_commit_count": invalid_count,
            "warning_commit_count": warning_commit_count,
            "issue_count": issue_count,
            "error_count": error_count,
            "warning_count": warning_count,
        },
        "results": [
            {
                "source": result.source,
                "is_valid": result.is_valid,
                "issue_count": len(result.issues),
                "error_count": result.error_count,
                "warning_count": result.warning_count,
                "issues": [
                    {
                        "code": issue.code,
                        "message": issue.message,
                        "line_number": issue.line_number,
                        "severity": issue.severity,
                        "suggestion": issue.suggestion,
                    }
                    for issue in result.issues
                ],
            }
            for result in results
        ],
    }


def render_json(results: list[ValidationResult]) -> None:
    console.print_json(data=_serialize_results(results))


def _fail(message: str, output_format: OutputFormat = OutputFormat.TEXT, hint: str | None = None) -> None:
    if output_format == OutputFormat.JSON:
        payload = {
            "schema_version": JSON_SCHEMA_VERSION,
            "tool_version": _tool_version(),
            "status": "error",
            "error": {
                "message": message,
            },
        }
        if hint is not None:
            payload["error"]["hint"] = hint
        console.print_json(data=payload)
        raise typer.Exit(code=2)

    error_console.print(f"[red]error:[/red] {message}")
    if hint is not None:
        error_console.print(f"hint: {hint}")
    raise typer.Exit(code=2)


def _fail_git_error(error: GitError, output_format: OutputFormat) -> None:
    _fail(error.message, output_format=output_format, hint=error.hint)


@app.command()
def run(
    input_value: Annotated[
        str | None,
        typer.Argument(
            help=(
                "Auto-detected input source: file path, commit ref, or commit range. "
                "Ranges are detected by '..' or '...'."
            ),
        ),
    ] = None,
    file_path: Annotated[
        Path | None,
        typer.Option("--file", "-f", help="Read a commit message from a file."),
    ] = None,
    commit: Annotated[
        str | None,
        typer.Option("--commit", "-c", help="Read a commit message from a commit hash or ref."),
    ] = None,
    commit_range: Annotated[
        str | None,
        typer.Option(
            "--range",
            "-r",
            help="Read commit messages from a git revision range, for example HEAD~5..HEAD.",
        ),
    ] = None,
    config_path: Annotated[
        Path | None,
        typer.Option(
            "--config",
            help="Path to a configuration file. Supports .coauthorcheck.toml or pyproject.toml.",
        ),
    ] = None,
    output_format: Annotated[
        OutputFormat,
        typer.Option(
            "--format",
            help="Output format for validation results.",
            case_sensitive=False,
        ),
    ] = OutputFormat.TEXT,
) -> None:
    try:
        config = load_config(config_path=config_path)
        messages = load_messages(input_value, file_path, commit, commit_range)
    except ConfigError as error:
        _fail(str(error), output_format=output_format)
    except ValueError as error:
        _fail(str(error), output_format=output_format)
    except FileNotFoundError as error:
        _fail(f"file not found: {error.filename}", output_format=output_format)
    except GitError as error:
        _fail_git_error(error, output_format=output_format)
    except OSError as error:
        _fail(f"unable to read input: {error}", output_format=output_format)

    results = [
        validate_message(
            message.source,
            message.message,
            config=config,
            author_name=message.author_name,
            author_email=message.author_email,
        )
        for message in messages
    ]

    if output_format == OutputFormat.JSON:
        render_json(results)
    else:
        for result in results:
            render_result(result)

        render_summary(results)

    if any(result.error_count > 0 for result in results):
        raise typer.Exit(code=1)


def main(argv: Sequence[str] | None = None) -> int:
    try:
        result = app(args=list(argv) if argv is not None else None, standalone_mode=False)
    except (typer.Exit, click.exceptions.Exit) as error:
        return int(error.exit_code)
    return int(result) if isinstance(result, int) else 0
